// I2C-Treiber   Dr. Weisser 14.10.2016     20.10.2019
//***************************************************************************
//  File........: i2c.c
//  Target(s)...: ATmega328p bzw. ATmega2560
//  Compiler....: AVR-GCC
//  Description.: I2C routines
//      bitbanging: geht fuer beliebige Pins, i2c_delay beachten
//      per AVR-Hardware: geht nur fuer bestimmte Pins an AVR-chips
//***************************************************************************
/* 14.10.2016 Uebernahme I2C-Code von Chan, Ueberarbeitung
              Trennung von I2C und RTC-Funktionen
              Pins flexibler per defines ansprechbar
   18.10.2016 435kHz klappt bei RTC DS3231
   15.11.2016 neue Funktionsnamen:
              i2c_write -> i2c_write_block
              i2c_read  -> i2c_read_block
              i2c_send  -> i2c_write
              i2c_rcvr  -> i2c_read
              Pinbezeichnung geaendert:
              SCL_LOW() -> SCL_L()
              SCL_HIGH()-> SCL_H()
              SDA_LOW() -> SDA_L()
              SDA_HIGH()-> SDA_H()
              SCL -> SCL_PIN
              SDA -> SDA_PIN
              Einbau optionale TWI-Unterstuetzung
   17.11.2016 MT_SLA_ACK dem Compiler nicht bekannt
              -> durch Festcode ersetzt, compiliert fehlerfrei
              LCD-Display klappt nicht mit TWI oder bitbanging
              I2C schreiben klappt denn LED-Licht geht aus
              klappt gut mit TWI, 100kHz und 400kHz ok
    5.12.2016 Test mit DS3231: ohne TWI ok.
              Test mit TWI klappt nicht - egal ob 100kHz od. 400kHz
              Schreiben LCD mit Einzelbefehlen klappt
   11.12.2016 Status bei TWI-Befehlen intensiver ausgewertet
              Fehler behoben. Auslesen RTC klappt mit/ohne TWI
   23.12.2018 ATmega2560 vorgesehen -> define dazu 
    16.3.2019 Pindefinitionen nach i2c.h verschoben 
   20.10.2019 Debug-Ausgaben outs(USART,PSTR) -> outs_p(USART,PSTR) 

*/

#include <stdint.h>         // Integer-Datentypen 8-64bit
#include <avr/io.h>         // Pins und Interrupt-Tabelle -> siehe auch iom328p.h
#include "main.h"           // CPU-Frequenz, sbi, cbi
#include "delay.h"          // Wartezeiten 2us, ms
#include "i2c.h"            // Funktionsprototypen
#include "usart.h"          // USART-Funktionen
#include "out.h"            // Ausgabefunktionen
#include <avr/pgmspace.h>   // PSTR



// I2C init
void i2c_init(void)  // init. IIC-Bus-Pins
{
   // set SCL/SDA as hi-Z
   // SCL/SDA-Pins auf Input gesetzt -> hi-Z
   //DDRC &= ~(_BV(SCL)|_BV(SDA));
   SCL_H();     // SCL auf Input
   SDA_H();     // SDA auf Input

   // SCL/SDA-Pins auf low (kein Pullup aktiv)
   //PORTC &= ~(_BV(SCL)|_BV(SDA));
   cbi(SCL_PORT,SCL_PIN);
   cbi(SDA_PORT,SDA_PIN);

#if TWI // TWI-Hardware

   // I2C-Bus-Geschwindigkeit setzen lt. Datenblatt S.222
   // SCL freq = F_CPU/(16+2*TWBR*4^TWPS)
   // set TWPS to zero
   cbi(TWSR,TWPS0);  // sollte 0 sein =default
   cbi(TWSR,TWPS1);  // sollte 0 sein =default
   TWBR = 12;  // 400kHz bei 16MHz Takt ***************************
   //TWBR = 72;  // 100kHz bei 16MHz Takt

#endif // TWI
}



#if TWI // TWI-Hardware
#else // nur bei bitbanging
// I2C bus delay
void i2c_delay(void)  // Delay fuer max. 400kHz mit bitbanging
{
   unsigned char n;

   for (n = 4; n; n--) // Original von Chan n=4
                        // n=4: 700ns = 5*125ns
                        // n=8: 1125ns = 9*125ns
                        // n=12: koennte fuer 100kHz passen
                        // n=8: 6us fuer SCL-Zyklus=167kHz
                        // n=6: 5us fuer SCL-Zyklus=200kHz
                        // n=4: 2.3us fuer SCL-Zyklus=435kHz
       PINB;
}
#endif // TWI



// generate start condition on I2C bus
void i2c_start(void)  // Start-Zeichen auf Bus
{
#if TWI // TWI-Hardware
   unsigned char status;

   // writing 1 to TWINT clears it, TWSTA=Start, TWEN=TWI-enable
   // Datenblatt S.226
   TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN);

   while (!(TWCR & (1<<TWINT))); // warten bis Start STA gesendet wurde

   // ggf. TWSR testen ob Start bzw. Repeated Start gesendet wurde
   status = TWSR & 0xF8;  // mask prescaler bits.
                          // TWI_START     0x08 =8
                          // TWI_REP_START 0x10 =16
   //if ((status != TWI_START) && (status != TWI_REP_START)) return FALSE;

#if DEBUG
   outs_p(USART,PSTR("\ni2c_start: Status ")); // ausgeben
   out_u08(USART,status,3,0);  // Ausgabe
#endif // DEBUG

#else
   // I2C-Start per bitbanging erzeugt
   SDA_H();  // SDA hi 125ns
   i2c_delay();
   SCL_H();
   i2c_delay();
   SDA_L();
   i2c_delay();
   SCL_L();
   i2c_delay();

#endif // TWI
}



// generate stop condition on I2C bus
void i2c_stop(void)  // Stop-Zeichen auf Bus
{
#if TWI // TWI-Hardware
   // writing 1 to TWINT clears it, TWSTO=Stop, TWEN=TWI-enable
   // Datenblatt S.226
   TWCR = (1<<TWINT) | (1<<TWSTO) | (1<<TWEN);

   // wait until stop condition is executed and bus released
   while (TWCR & (1<<TWINT));  // solange TWINT gesetzt

#else
   // I2C-Stop per bitbanging erzeugt
   SDA_L();
   i2c_delay();
   SCL_H();
   i2c_delay();
   SDA_H();
   i2c_delay();

#endif // TWI
}



// write byte to I2C bus
unsigned char i2c_write(unsigned char data)  // Byte schicken auf Bus
                        // Datenbyte oder
                        // Slave-Adresse + R/W-bit
{
#if TWI // TWI-Hardware
   unsigned char status;

   // TWDR Datenregister: Datenblatt S.224,226
   TWDR = data;                   // TWDR contains byte to send
   TWCR = (1<<TWINT) | (1<<TWEN); // send content of TWDR

   while (!(TWCR & (1<<TWINT)));  // warten bis byte verschickt

#if DEBUG
   outs_p(USART,PSTR("\ni2c_write: Data ")); // ausgeben
   out_u08(USART,data,3,0);  // Ausgabe
#endif // DEBUG

   // TWI Status Register auswerten
   // TWSR gibt Status an in bits TWS3-7:
   //        TWPS1  = 0 Prescaler =1  S.243
   //         TWPS0 = 0
   // TWS7
   //  TWS6
   //   TWS5
   //    TWS4
   //      TWS3
   // 1111 1000
   status = TWSR & 0xF8;    // mask prescaler bits
                            // 0xF8 = 1111 1000 schaut nur TWS3-7 an
                            // Abfrage auf MT_SLA_ACK klappt nicht
                            // lt S.228 Codes 0x18, 0x20, 0x38 moeglich
                            // vgl. defines in i2c.h
                            // TWI_MTX_DATA_ACK  0x28 =40
                            // TWI_MTX_ADR_ACK   0x18 =24
                            // TWI_MRX_ADR_ACK   0x40 =64
   if ( (status == TWI_MTX_DATA_ACK) |  // data ACK Senden ok
        (status == TWI_MTX_ADR_ACK) |   // adr ACK Senden ok
        (status == TWI_MRX_ADR_ACK) ){  // adr ACK Empfangen ok

#if DEBUG
      outs_p(USART,PSTR("\ni2c_write: Status ACK ")); // ausgeben
      out_u08(USART,status,3,0);  // Ausgabe
#endif // DEBUG

      // bei erfolgreich verschickt 1 rueckgeben, bei Fehler 0 rueckgeben
      return 1;  // ACK da, kein Fehler
   }

#if DEBUG
   outs_p(USART,PSTR("\ni2c_write: Status NACK ")); // ausgeben
   out_u08(USART,status,3,0);  // Ausgabe
#endif // DEBUG

   return 0;     // Fehler, kein ACK vom Slave

#else

   unsigned char byte = 0x80;
   unsigned char ack;

#if DEBUG
   outs_p(USART,PSTR("\ni2c_write: Data ")); // ausgeben
   out_u08(USART,data,3,0);  // Ausgabe
#endif // DEBUG

   // 8 bits Daten auf I2C-Bus schieben
   // mit oberstem bit (MSB) anfangen
   do {
      if (data & byte)  // bit gesetzt?
         SDA_H();    // falls ja -> SDA high setzen
      else
         SDA_L();    // sonst -> SDA low setzen

      // SDA ausgeben -> dazu SCL takten
      i2c_delay();
      SCL_H();
      i2c_delay();
      SCL_L();
      i2c_delay();

   } while (byte >>= 1);

   // 9. bit erzeugen und abfragen -> hat Slave empfangen? ACK da?
   SDA_H();
   i2c_delay();
   SCL_H();
   ack = SDA_VAL ? 0 : 1;  // sample ACK
   i2c_delay();
   SCL_L();
   i2c_delay();

#if DEBUG
   outs_p(USART,PSTR("\ni2c_write: Status ack ")); // ausgeben
   out_u08(USART,ack,3,0);  // Ausgabe
#endif // DEBUG

   // ACK zurueckgeben
   return ack;  // 1=ok, 0=Fehler (kein ACK vom Slave)

#endif // TWI
}



// read byte from I2C bus
unsigned char i2c_read(unsigned char ack)  // Byte empfangen ueber Bus
{
#if TWI // TWI-Hardware

#if DEBUG
   outs_p(USART,PSTR("\ni2c_read: ack ")); // ausgeben
   out_u08(USART,ack,3,0);  // Ausgabe
#endif // DEBUG

   // TWEA = enable/disable ACK: Datenblatt S.224,226
   if (ack)  // soll ACK beim Lesen gesetzt werden?
      TWCR = (1<<TWINT) | (1<<TWEA) | (1<<TWEN); // ACK setzen
   else
      TWCR = (1<<TWINT) | (0<<TWEA) | (1<<TWEN); // ACK nicht setzen

   while (!(TWCR & (1<<TWINT)));  // warte bis Byte empfangen wurde

#if DEBUG
   outs_p(USART,PSTR("\ni2c_read: TWDR ")); // ausgeben
   out_u08(USART,TWDR,3,0);  // Ausgabe
#endif // DEBUG

   // gibt das empfangene Byte zurueck
   return TWDR;

#else

   unsigned int byte = 1;  // ganz rechtes bit gesetzt

   // 8 bits einlesen und in byte speichern
   do {
      byte <<= 1;  // bits eine Stelle nach links geschoben
      SCL_H();
      if (SDA_VAL)
         byte++;
      i2c_delay();
      SCL_L();
      i2c_delay();
   } while (byte < 0x100);  // solange bis die 1 8x links geschoben

   // 9. bit = ACK oder NACK
   // ACK wie gewuenscht setzen: letztes Byte kein ACK !
   if (ack)    // soll ACK gesetzt werden?
      SDA_L(); // ja -> ACK setzen
   else
      SDA_H(); // nein -> ACK nicht setzen
   i2c_delay();
   SCL_H();
   i2c_delay();
   SCL_L();
   SDA_H();
   i2c_delay();

   return (unsigned char)byte;

#endif // TWI
}



// read I2C block
unsigned char i2c_read_block(   // blockweise lesen vom Bus
   unsigned char dev,   // device address
   unsigned char adr,   // read start address
   unsigned int cnt,    // read byte count (auch >255)
   void* buff)          // read data buffer
{
   unsigned char *rbuff = buff;
   unsigned char n;

   if (!cnt) return 0;

#if DEBUG
   outs_p(USART,PSTR("\ni2c_read_block: dev ")); // ausgeben
   out_u08(USART,dev,3,0);  // Ausgabe
#endif // DEBUG

   n = 10;
   do {    // select device
      i2c_start();
   } while (!i2c_write(dev) && --n);

   if (n){
      if (i2c_write(adr)){       // set start address
         i2c_start();            // reselect device in read mode
         if (i2c_write(dev | 1)){
            do {                 // receive data
               cnt--;
               *rbuff++ = i2c_read(cnt ? 1 : 0);
            } while (cnt);
         }
      }
   }

   i2c_stop();  // deselect device

   return cnt ? 0 : 1;
}



// write I2C block
unsigned char i2c_write_block(  // blockweise schreiben
   unsigned char dev,   // device address
   unsigned char adr,   // write start address
   unsigned int cnt,    // write byte count (auch >255)
   const void* buff)    // data to be written
{
   const unsigned char *wbuff = buff;
   unsigned char n;

   if (!cnt) return 0;

   n = 10;
   do {           // select device
      i2c_start();
   } while (!i2c_write(dev) && --n);

   if (n){
      if (i2c_write(adr)){ // set start address
         do {              // send data
            if (!i2c_write(*wbuff++))
               break;
         } while (--cnt);
      }
   }

   i2c_stop();    // deselect device

   return cnt ? 0 : 1;
}
