// USART-Treiber   Dr. Weisser 21.11.2016  30.6.2019
//***************************************************************************
//  File........: usart.c
//  Target(s)...: ATmega328p, ATmega2560, Atmega32u2, Atmega32u4
//  Compiler....: AVR-GCC
//  Description.: AVR USART routines
//***************************************************************************
/* 21.11.2016 Umformatierung auf Einrueckung 3
              Aenderung der Funktionsnamen:
                USART_init     -> usart_init
                USART_Tx       -> usart_putc
                USART_Tx_hex   -> usart_putc_hex
                USART_Tx_msg   -> usart_puts
                USART_Tx_byte  -> usart_put_u08_2
                USART_Tx_int   -> usart_put_u16_5
                USART_Rx       -> usart_getc
                USART_getValue -> usart_get_u08_2
              Funktionen leicht ueberarbeitet
              neue Ausgabefunktionen, noch nicht aktiv:
                usart_put_u16
                usart_put_u16f
                usart_put_u32
                usart_put_u32f
   24.11.2016 Funktionen usart_put_u16, usart_put_u32 verzichtbar
              -> stattdessen usart_put_u16f, usart_put_u32f
              umbenannt in usart_put_u16, usart_put_u32
              Funktionen kombiniert mit LCD-Ausgabe:
                put_u16, put_u32
              universelle Funktionen - kombiniert mit LCD
                putc, puts -> Meldung conflicting types
              offenbar sind solche Funktionen bereits vorhanden
              Funktionen ausgelagert:
                putc    -> outc in out.c
                puts    -> outs in out.c
                put_u16 -> out_u16 in out.c
                put_u32 -> out_u32 in out.c
                put_u08 -> out_u08 in out.c
    23.2.2019 Globale Variable load_state fuer Laststeuerung 
              Belegung von load_state per USART 
    26.4.2019 Umschaltung serieller Port zwischen atmega328p, 2560, 32u2 
              Globale Variable load_state fuer Laststeuerung auskommentiert
    24.6.2019 atmega32u4 hinzu bei Umschaltung USART
    30.6.2019 Umschaltung fuer chips veraendert -> Uebernahme aus Makefile 

*/

#include <stdint.h>        // Integer-Datentypen 8-64bit
#include <avr/io.h>        // Pins und Interrupt-Tabelle -> siehe auch iom328p.h
#include "main.h"          // CPU-Frequenz, sbi, cbi
#include "usart.h"         // USART-Funktionen
#include "out.h"           // Ausgabe-Funktionen Zeichen, Zahlen
//#include "bcd.h"         // ggf. BCD-Umwandlungen fuer Ausgabe
#include <avr/interrupt.h> // Interrupts
#include <avr/pgmspace.h>  // PSTR

// ggf. define in usart.h
//#define __AVR_ATmega32U4__ 1 // Hardwarewahl: 32U4 (=1) oder nicht (=0)



// Global variables
//extern volatile unsigned char load_state;  // Zustandsvariable fuer Laststeuerung in main.c


/*****************************************************************************
*   Function name : usart_init
*   Returns :       none
*   Parameters :    unsigned int baudrate
*   Purpose :       initialize USART
*****************************************************************************/
void usart_init(unsigned int baudrate)
{
   // set baud rate: Table 19-12. Datenblatt ATmega328P S.203
            // bei 16MHz U2X0=1 38.4k -> 51 Fehler 0.2%
            //                  57.6k    34        0.8%
            //                  115.2k   16        2.1%
            //                  250k      7        0%
            //                  0.5M      3        0%
            //                  1M        1        0%
#if defined(__AVR_ATmega328P__) || defined (__AVR_ATmega2560__) 
                             // Atmega328/Arduino nano3 oder
                             // Atmega2560 
   // nutze USART0
   // set baud rate
   UBRR0H = (unsigned char)(baudrate>>8);  // erst high, dann low
   UBRR0L = (unsigned char)baudrate;       // baudrate gesetzt
   
   // enable 2x speed
   UCSR0A = (1<<U2X0);
   
   // enable receiver and transmitter
   UCSR0B = (1<<RXCIE0)|(0<<TXCIE0)|(0<<UDRIE0)|(1<<RXEN0)|(1<<TXEN0); // mit Interrupt bei Empfang
   
   // async. mode, kein parity, 1 Stop-Bit, 8bit, async
   UCSR0C = (0<<UMSEL01)|(0<<UMSEL00)|(0<<UPM01)|(0<<UPM00)|(0<<USBS0)|(1<<UCSZ01)|(1<<UCSZ00)|(0<<UCPOL0);

#elif defined(__AVR_ATmega32U2__) || defined(__AVR_ATmega32U4__) || defined(__AVR_AT90USB1286__) 
                             // Atmega32U2 oder 
                             // Atmega32U4 oder 
                             // AT90USB1286 
   // nutze USART1
   // set baud rate
   UBRR1H = (unsigned char)(baudrate>>8);  // erst high, dann low 
   UBRR1L = (unsigned char)baudrate;       // baudrate gesetzt
   
   // enable 2x speed
   UCSR1A = (1<<U2X1);
   
   // enable receiver and transmitter
   UCSR1B = (1<<RXCIE1)|(0<<TXCIE1)|(0<<UDRIE1)|(1<<RXEN1)|(1<<TXEN1); // mit Interrupt bei Empfang
   
   // async. mode, kein parity, 1 Stop-Bit, 8bit, async
   UCSR1C = (0<<UMSEL11)|(0<<UMSEL10)|(0<<UPM11)|(0<<UPM10)|(0<<USBS1)|(1<<UCSZ11)|(1<<UCSZ10)|(0<<UCPOL1);

#endif // Ende Hardwareauswahl 

}



/*****************************************************************************
*   Function name : usart_putc
*   Returns :       none
*   Parameters :    char data: byte to send
*   Purpose :       send one byte through USART
*                   ein Zeichen (1 Byte) seriell schicken
*****************************************************************************/
// siehe outc(unsigned char device, unsigned char data)



/*****************************************************************************
*   Function name : usart_puts
*   Returns :       nothing
*   Parameters :    pointer to program memory string
*   Purpose :       sends string out on UART
*****************************************************************************/
// siehe outs(unsigned char device, const char *s)



/*****************************************************************************
*   Function name : usart_putc_hex
*   Returns :       none
*   Parameters :    char data: byte to send
*   Purpose :       sends one byte through USART as a two char Hex number
*                   Byte seriell schicken als 2 Hex-Ziffern
*                   z.B. 0x00 =0, 0x08 =8, 0x0a =9 0xff =255
*                   der GCC wird mit -funsigned-char genutzt -> unsigned char !
*
*                   z.B.  0 = 0000 0000 als 00 ausgeben -> 30 und 30 senden (zwei ASCII-Nullen)
*                         9 = 0000 1001 als 09 ausgeben -> 30 und 39 senden
*                        10 = 0000 1010 als 0A ausgeben -> 30 und 'A'=41 senden
*                        15 = 0000 1111 als 0F ausgeben
*                       255 = 1111 1111 als FF ausgeben
*****************************************************************************/
void usart_putc_hex(unsigned char data)
{
   unsigned char data_l, data_h;

   data_h = data >> 4;    // 8 bit-Wert 4 bit nach rechts schieben = high-Teil
   data_h += '0';         // 30 addieren setzt linke 4 bit auf 0011=3

   if (data_h > '9')
      data_h += 7;        // bei Ziffer >9 (z.B.10=3A) 7 addieren (aus 3A wird 41=A)

   data_l = data & 0x0f;  // unteren Teil der 8bit auf 0 setzen
   data_l += '0';         // 30 addieren

   if (data_l > '9')
      data_l += 7;        // >39 (z.B 3A) 7 addieren (z.B. 3B,3C,3D,3E,3F,40,41=A)

   outc(USART,data_h);    // schicke erst High-Byte, dann Low-Byte
   outc(USART,data_l);
}



/*****************************************************************************
*   Function name : usart_put_u08_2 (2 digits)
*   Returns :       none
*   Parameters :    unsigned char data: byte to send
*   Purpose :       sends one byte through USART as a 2 digit decimal number
*                   ein Byte als 2 Dezimalziffern senden (nur 0-99 zulssig)
*                     0 (=0x00) als 0 0 senden
*                     1 (=0x01) als 0 1 senden
*                     9 (=0x09) als 0 9 senden
*                    10 (=0x0A) als 1 0 senden
*                    15 (=0x0F) als 1 5 senden
*                    16 (=0x10) als 1 6 senden
*                    99 (=0x63) als 9 9 senden
*                   100 (=0x64) als A 0 senden (ab hier wird es unsinnig, daher verboten !)
*                   101 (=0x65) als A 1 senden?
*                   109 (=0x6D) als A 9 senden?
*                   110 (=0x6E) als B 0 senden?
*                   120         als C 0 senden?
*                   130         als D 0 senden?
*                   140         als E 0 senden?
*                   150         als F 0 senden?
*                   151         als F 1 senden?
*                   159         als F 9 senden?
*                   255 (=0xFF) als ???
*
*****************************************************************************/
/*
void usart_put_u08_2(unsigned char data)
{
   usart_putc_hex(char_to_bcd2(data));  // benoetigt bcd.c
}
*/


/*****************************************************************************
*   Function name : usart_put_u16_5 (5 digits)
*   Returns :       none
*   Parameters :    unsigned int data: 10bit value to send
*   Purpose :       sends value through USART as a 4 digit decimal number
*                   Wert als 4 Dezimalziffern senden (nur 0-1023 zulssig)
*                     0 (=0x00) als 0 0 0 0 senden
*                     1 (=0x01) als 0 0 0 1 senden
*                     9 (=0x09) als 0 0 0 9 senden
*                    10 (=0x0A) als 0 0 1 0 senden
*                    15 (=0x0F) als 0 0 1 5 senden
*                    16 (=0x10) als 0 0 1 6 senden
*                    99 (=0x63) als 0 0 9 9 senden
*                   100 (=0x64) als 0 1 0 0 senden
*                  1000         als 1 0 0 0 senden
*                  1023         als 1 0 2 3 senden
*                  9999         als 9 9 9 9 senden
*                 10000         als A 0 0 0 senden? (ab hier wird es unsinnig, daher verboten !)
*                 65535         als F F F F senden?
*****************************************************************************/
/*
void usart_put_u16_5(unsigned int data)
{
   unsigned long temp;
   unsigned char tmp;
   unsigned char ziff=0; // fuehrende Null da? fuer Unterdrueckung
   signed char i;  // muss signed sein, damit for-Schleife beendet wird mit <0 !

   temp = int_to_bcd5(data);  // xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx 2 Integer a 16bit
                              //   0    0    0    0    0    0    0    0
                              //   0    0    0    6    5    5    3    5

   //USART_Tx ((( (unsigned char) (temp>>16) )&0x0f)+'0'); // soll eine ASCII 6 ausgeben
   //USART_Tx ((( (unsigned char) (temp>>12) )&0x0f)+'0');
   //USART_Tx ((( (unsigned char) (temp>>8)  )&0x0f)+'0');
   //USART_Tx ((( (unsigned char) (temp>>4)  )&0x0f)+'0');
   //USART_Tx ((( (unsigned char) (temp>>0)  )&0x0f)+'0');

   for (i=16; i>=0; i-=4){
      // i=16 0x00065535 -> 0x00000006 -> 0x06 -> 0x06 -> 36 ausgeben
      // i=12 0x00065535 -> 0x00000065 -> 0x65 -> 0x05 -> 35 ausgeben
      // i=8  0x00065535 -> 0x00000655 -> 0x55 -> 0x05 -> 35 ausgeben
      // i=4  0x00065535 -> 0x00006553 -> 0x53 -> 0x03 -> 33 ausgeben
      // i=0  0x00065535 -> 0x00065535 -> 0x35 -> 0x05 -> 35 ausgeben

      tmp= (((unsigned char) (temp>>i)) &0x0f)+'0'; // einzelnes Zeichen: 30 addieren fr ASCII-Ausgabe

      if (tmp=='0') {     // eine Ziffer = Null da?
         if (ziff==0) {     // ist es eine fuehrende 0?
            if (i==0) {     // ist die letzte Ziffer eine 0?
               usart_putc(tmp); // diese Null ausgeben
            }
            else {        // nein, also unterdruecken
               usart_putc(' '); // fuehrende Null als Space ausgeben
            }
         }
         else {
            usart_putc(tmp); // Nullen in der Zahl als Nullen ausgeben
         }
      }
      else {              // eine Ziffer ungleich Null da
         ziff=1;
         usart_putc(tmp); // Ziffer 0-9 als ASCII ausgeben
      }
   }
}
*/


/*****************************************************************************
*   Function name : usart_getc
*   Returns :       unsigned char: byte received
*   Parameters :    none
*   Purpose :       receives one byte from USART
*                   ein Byte seriell empfangen
*****************************************************************************/
unsigned char usart_getc(void)
{
#if defined(__AVR_ATmega328P__) || defined (__AVR_ATmega2560__) 
                             // Atmega328/Arduino nano3 oder
                             // Atmega2560 
   // nutze USART0 
   while (!(UCSR0A & (1<<RXC0)));
   return UDR0;  // Datenbyte empfangen, wenn erlaubt

#elif defined(__AVR_ATmega32U2__) || defined(__AVR_ATmega32U4__) || defined(__AVR_AT90USB1286__) 
                             // Atmega32U2 oder 
                             // Atmega32U4 oder 
                             // AT90USB1286 
   // nutze USART1 
   while (!(UCSR1A & (1<<RXC1)));
   return UDR1;  // Datenbyte empfangen, wenn erlaubt

#endif // Hardwareauswahl 

}



/*****************************************************************************
*   Function name : usb_serial_getc
*   Returns :       unsigned char: byte received
*   Parameters :    none
*   Purpose :       receives one byte from USB
*                   ein Byte seriell von USB empfangen
*****************************************************************************/
// nutze dazu usb_serial_getchar(); aus usb_serial.c 
// usb_serial_getchar() liefert jedoch 16bit statt 8bit ! Vorsicht ! 



/****************************************************************************
*   Function Name : ISR_USART_RX
*   Returns :       none
*   Parameters :    none
*   Purpose :       Interrupt service routine for data received from UART
*                   receives one byte of data from USART. Determines
*                   command char received and execute code as required.
*   vgl. WinAVR-20100110/doc/avr-libc/avr-libc-user-manual/group__avr__interrupts.html
*   Eintraege zu ATmega328P
*****************************************************************************/

#if defined(__AVR_ATmega328P__) 
                             // Atmega328/Arduino nano3 
// nutze USART 
ISR(USART_RX_vect)

#elif defined (__AVR_ATmega2560__) 
                             // Atmega2560 
// nutze USART0 
ISR(USART0_RX_vect)

#elif defined (__AVR_ATmega32U2__) || defined (__AVR_ATmega32U4__) 
                             // Atmega32U2 oder 
                             // Atmega32U4 
// nutze USART1 
ISR(USART1_RX_vect)

#endif // Ende Hardwareauswahl
{
   unsigned char receive;
   //unsigned int i;
   //unsigned int wert;
   //unsigned int rambuffer[500]; // Werte im RAM

   receive = usart_getc(); // get command character from UART receive register
   outc(USART,receive);    // empfangenes Zeichen zur Bestaetigung ruecksenden
   outc(USART,'\n');       // neue Zeile

/*
   // Last-Messung steuern
   if (receive=='S'){      // Last-Messung starten
      cli(); // Befehl abarbeiten -> weitere Interrupts solange sperren
      load_state='S'; 
      sei(); // Interrupts wieder freigeben
   }
   else if (receive=='E'){ // Last-Messung beenden
      cli(); // Befehl abarbeiten -> weitere Interrupts solange sperren
      load_state='E'; 
      sei(); // Interrupts wieder freigeben
   }
   else if (receive=='R'){ // erfasste Daten aus SRAM auslesen
      cli(); // Befehl abarbeiten -> weitere Interrupts solange sperren
      load_state='R'; 
      sei(); // Interrupts wieder freigeben
   }
   else if (receive=='D'){ // Daten im SRAM loeschen
      cli(); // Befehl abarbeiten -> weitere Interrupts solange sperren
      load_state='D'; 
      sei(); // Interrupts wieder freigeben
   } */
}



/*****************************************************************************
*   Function name : usart_get_u08_2
*   Returns :       unsigned char: 0 for unsuccessful / 1 for successful
*   Parameters :    unsigned char
*   Purpose :       takes input from user via Uart, does basic error check
*****************************************************************************/
/*
unsigned char usart_get_u08_2(unsigned char *value, unsigned char lowerbound, unsigned char upperbound)
{
   unsigned char retval;   // function return value
   unsigned char temp;
   retval = 1;    // set retval to 1 as default (successful)
   *value = 0;    // initialise the parameter

   // print out the limits for the input
   out_u08(USART,lowerbound);
   outc(USART,'-');
   out_u08(USART,upperbound);
   outc(USART,' ');

   temp = usart_getc();   // get first character sent
   *value = ((char)(temp - '0')) * 10; // convert char from Ascii to Hex and multiply by 10

   temp = usart_getc();   // get second character sent
   *value += ((char)(temp - '0'));  // convert char from Ascii to Hex and add to yearValue

   if ((*value < lowerbound) | (*value > upperbound)){
      // check the value provided by user is valid
      // if not then increment error counter return 0
      retval = 0;       // set retval to 0 signifying unsuccessful operation
      outs_p(USART,PSTR("ERR:"));
   }
   return retval;
}
*/
